#version 330

#define MAX_LIGHTS 10

uniform mat4 model;
uniform mat4 viewMatrix;
uniform sampler2D tex;
uniform int numLights;
uniform vec3 ambientLight;
uniform float specularFactor;
uniform float specularAlpha;
uniform vec3 cameraPosition;
uniform vec3 fogColor;
uniform bool fogEnabled;

in vec3 fragPosition;
in vec3 fragNormal;
in vec2 fragUV;

out vec4 fragColor;
	
uniform struct Light {
	int type;
	bool enabled;
	vec3 intensity;
	vec3 position;
	float constAttenuation;
	float linearAttenuation;
	float quadraticAttenuation;
	vec3 coneDirection;
	float coneInnerAngle;
	float coneOuterAngle;
} lights[MAX_LIGHTS];


vec3 getDiffuseLightIntensity(vec3 position, vec3 normal, Light light);
float getFogFactor();

void main()
{
	// Calculate normal in world coordinates
    mat3 normalMatrix = transpose(inverse(mat3(model)));
    vec3 normal = normalize(normalMatrix * fragNormal);
    
    // Calculate the location of this fragment (pixel) in world coordinates
    vec3 position = vec3(model * vec4(fragPosition, 1));
	
	// Get texture color
	vec4 surfaceColor = vec4(texture(tex, fragUV).rgb, 1);

	vec3 allIntensities = ambientLight;
	vec3 surfaceToLight = vec3(0);
	float brightness = 0;
    
	for(int i = 0; i < numLights; i++) 
	{
		if(!lights[i].enabled) continue;

		allIntensities += getDiffuseLightIntensity(position, normal, lights[i]);
	}	

	// Calculate final color dependend on texture color and all lights
    fragColor = vec4(allIntensities * surfaceColor.rgb, surfaceColor.a);

	// Mix with fog (LERP)
	float fogFactor = 1.0f;
	if(fogEnabled) fogFactor = getFogFactor();
	fragColor = vec4(fogColor, 1) * (1 - fogFactor) + fragColor * fogFactor;
}

float getFogFactor()
{
	// Linear fog
	vec4 viewPos = viewMatrix * model * vec4(fragPosition, 1.0f);
	float dist = min(60.0f, abs(viewPos.z));
	float fogFactor = 1.0f / exp(pow(dist * 0.02f, 2));
	fogFactor = clamp(fogFactor, 0.0f, 1.0f);

	return fogFactor;
}

vec3 getDiffuseLightIntensity(vec3 position, vec3 normal, Light light) 
{
	vec3 surfaceToLight = vec3(0);
	float brightness = 0;
	float attenuation = 0;

	switch(light.type) 
	{
		case 0: // POINT LIGHTS
			// Calculate the vector from this pixels surface to the light source
			surfaceToLight = light.position - position;

			// Calculate the cosine of the angle of incidence
			brightness = dot(normal, surfaceToLight) / (length(surfaceToLight) * length(normal));
			brightness = clamp(brightness, 0, 1);

			// Calculate attenuation
			float d = length(surfaceToLight);
			attenuation = light.constAttenuation + d * light.linearAttenuation + (d * d) * light.quadraticAttenuation;

			// Calculate intensity of pixel for current light
			return (light.intensity / attenuation) * brightness;

		case 1: // DIRECTIONAL LIGHTS
			// Calculate the vector from this pixels surface to the light source
			surfaceToLight = normalize(light.position) * -1;

			// Calculate the cosine of the angle of incidence
			brightness = dot(normal, surfaceToLight) / (length(surfaceToLight) * length(normal));
			brightness = clamp(brightness, 0, 1);
				
			// Calculate intensity of pixel for current light
			return light.intensity * brightness;

			break;

		case 2: // SPOT LIGHTS        
			// Calculate the vector from this pixels surface to the light source
			surfaceToLight = light.position - position;

			float lightToSurfaceAngle = dot(-normalize(surfaceToLight), normalize(light.coneDirection));

			if(lightToSurfaceAngle > light.coneOuterAngle) { // in light cone

				// Calculate the cosine of the angle of incidence
				brightness = dot(normal, surfaceToLight) / (length(surfaceToLight) * length(normal));
				brightness = clamp(brightness, 0, 1);

				// Calculate attenuation
				float d = length(surfaceToLight);
				attenuation = light.constAttenuation + d * light.linearAttenuation + (d * d) * light.quadraticAttenuation;
					
				// Calculate intensity of pixel for current light
				vec3 intensity = (light.intensity / attenuation) * brightness;

				if(lightToSurfaceAngle <= light.coneInnerAngle) {
					// Lerp between inner and outer cone
					intensity = intensity * ((lightToSurfaceAngle - light.coneOuterAngle) / (light.coneInnerAngle - light.coneOuterAngle));
				}

				return intensity;
			}

			return vec3(0);
	} 
}